# -*- coding: utf-8 -*-
# Copyright 2017 Joshua Bronson. All Rights Reserved.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
Property-based tests using https://hypothesis.readthedocs.io
"""

from collections import OrderedDict
from os import getenv

import pytest
from hypothesis import assume, given, settings
from hypothesis.strategies import integers, lists, tuples
from bidict import (
    IGNORE, OVERWRITE, RAISE,
    bidict, OrderedBidict,
    frozenbidict, FrozenOrderedBidict)
from bidict.compat import iteritems


settings.register_profile('default', settings(max_examples=200, deadline=None))
settings.load_profile(getenv('HYPOTHESIS_PROFILE', 'default'))


# pylint: disable=C0111
def to_inv_odict(items):
    return OrderedDict((v, k) for (k, v) in items)


def dedup(items):
    pruned = list(iteritems(to_inv_odict(iteritems(to_inv_odict(items)))))
    assume(len(pruned) >= len(items) // 2)
    return pruned


# pylint: disable=C0103
ondupbehaviors = (IGNORE, OVERWRITE, RAISE)
mutable_bidict_types = (bidict, OrderedBidict)
immutable_bidict_types = (frozenbidict, FrozenOrderedBidict)
bidict_types = mutable_bidict_types + immutable_bidict_types
mutating_methods_by_arity = {
    0: ('clear', 'popitem'),
    1: ('__delitem__', 'pop', 'setdefault', 'move_to_end'),
    2: ('__setitem__', 'pop', 'put', 'forceput', 'setdefault'),
    -1: ('update', 'forceupdate'),
}
immutable = integers()
itemlists = lists(tuples(immutable, immutable))
inititems = itemlists.map(dedup)


@pytest.mark.parametrize('B', bidict_types)
@given(init=inititems)
def test_equality(B, init):  # noqa
    b = B(init)
    d = dict(init)
    o = OrderedDict(init)
    oi = to_inv_odict(iteritems(o))
    di = OrderedDict(oi)
    assert b == d
    assert b == o
    assert not b != d
    assert not b != o
    assert b.inv == oi
    assert b.inv == di
    assert not b.inv != oi
    assert not b.inv != di


@pytest.mark.parametrize('B', bidict_types)
@given(init=inititems)
def test_bidirectional_mappings(B, init):  # noqa
    ordered = bool(getattr(B, '__reversed__', None))
    C = list if ordered else sorted  # noqa
    b = B(init)
    keysf = C(k for (k, v) in iteritems(b))
    keysi = C(b.inv[v] for (k, v) in iteritems(b))
    assert keysf == keysi
    valsf = C(b[k] for (v, k) in iteritems(b.inv))
    valsi = C(v for (v, k) in iteritems(b.inv))
    assert valsf == valsi


@pytest.mark.parametrize(
    'arity, methodname',
    [(a, m) for (a, ms) in iteritems(mutating_methods_by_arity) for m in ms])
@pytest.mark.parametrize('B', mutable_bidict_types)  # noqa
@given(init=inititems, arg1=immutable, arg2=immutable, items=itemlists)
def test_consistency_after_mutation(arity, methodname, B, init, arg1, arg2, items):
    method = getattr(B, methodname, None)
    if not method:
        return
    args = []
    if arity == -1:
        args.append(items)
    else:
        if arity > 0:
            args.append(arg1)
        if arity > 1:
            args.append(arg2)
    b0 = B(init)
    b1 = b0.copy()
    try:
        method(b1, *args)
    except Exception as exc:  # pylint: disable=W0703
        # method should fail clean, i.e. b1 should be in the same state it was before the call.
        assert b1 == b0, '%r did not fail clean: %r' % (method, exc)
        assert b1.inv == b0.inv, '%r did not fail clean: %r' % (method, exc)
    # Whether method failed or succeeded, b1 should pass consistency checks.
    assert len(b1) == sum(1 for _ in iteritems(b1))
    assert len(b1) == sum(1 for _ in iteritems(b1.inv))
    assert b1 == dict(iteritems(b1))
    assert b1.inv == dict(iteritems(b1.inv))
    assert b1 == to_inv_odict(iteritems(b1.inv))
    assert b1.inv == to_inv_odict(iteritems(b1))


@pytest.mark.parametrize('on_dup_key', ondupbehaviors)
@pytest.mark.parametrize('on_dup_val', ondupbehaviors)
@pytest.mark.parametrize('on_dup_kv', ondupbehaviors)
@pytest.mark.parametrize('B', mutable_bidict_types)
@given(init=inititems, items=itemlists)
def test_putall(on_dup_key, on_dup_val, on_dup_kv, B, init, items):  # noqa
    b0 = B(init)
    expect = b0.copy()
    expectexc = None
    for (k, v) in items:
        try:
            expect.put(k, v, on_dup_key=on_dup_key, on_dup_val=on_dup_val, on_dup_kv=on_dup_kv)
        except Exception as exc:  # pylint: disable=W0703
            expectexc = exc
            expect = b0  # bulk updates fail clean
            break
    check = b0.copy()
    checkexc = None
    try:
        check.putall(items, on_dup_key=on_dup_key, on_dup_val=on_dup_val, on_dup_kv=on_dup_kv)
    except Exception as exc:  # pylint: disable=W0703
        checkexc = exc
    assert type(checkexc) == type(expectexc)  # pylint: disable=C0123
    assert check == expect
    assert check.inv == expect.inv



# -*- coding: utf-8 -*-
# Copyright 2017 Joshua Bronson. All Rights Reserved.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""Implements :class:frozenbidict."""

from collections import ItemsView

from ._abc import BidirectionalMapping
from ._dup import RAISE, OVERWRITE, IGNORE
from ._exc import (
    DuplicationError, KeyDuplicationError, ValueDuplicationError, KeyAndValueDuplicationError)
from ._miss import _MISS
from .compat import PY2, iteritems
from .util import pairs


def _proxied(methodname, attrname='fwdm', doc=None):
    """Make a func that calls the indicated method on the indicated attribute."""
    def proxy(self, *args):
        """(__doc__ set dynamically below)"""
        attr = getattr(self, attrname)
        meth = getattr(attr, methodname)
        return meth(*args)
    proxy.__name__ = methodname
    proxy.__doc__ = doc or "Like dict's ``%s``." % methodname
    return proxy


# pylint: disable=invalid-name,too-many-instance-attributes
class frozenbidict(BidirectionalMapping):  # noqa: N801
    u"""
    Immutable, hashable bidict type.

    Also serves as a base class for the other bidict types.

    .. py:attribute:: fwd_cls

        The :class:`Mapping <collections.abc.Mapping>` type
        used for the backing :attr:`fwdm` mapping,
        Defaults to :class:`dict`.
        Override this if you need different behavior.

    .. py:attribute:: inv_cls

        The :class:`Mapping <collections.abc.Mapping>` type
        used for the backing :attr:`invm` mapping.
        Defaults to :class:`dict`.
        Override this if you need different behavior.

    .. py:attribute:: on_dup_key

        The default :class:`DuplicationPolicy` used in the event that an item
        duplicates only the key of another item,
        when a policy has not been specified explicitly
        (e.g. the policy used by :meth:`__setitem__` and :meth:`update`).
        Defaults to :class:`OVERWRITE <DuplicationPolicy.OVERWRITE>`
        to match :class:`dict`'s behavior.

    .. py:attribute:: on_dup_val

        The default :class:`DuplicationPolicy` used in the event that an item
        duplicates only the value of another item,
        when a policy has not been specified explicitly
        (e.g. the policy used by :meth:`__setitem__` and :meth:`update`).
        Defaults to :class:`RAISE <DuplicationPolicy.RAISE>`
        to prevent unintended overwrite of another item.

    .. py:attribute:: on_dup_kv

        The default :class:`DuplicationPolicy` used in the event that an item
        duplicates the key of another item and the value of yet another item,
        when a policy has not been specified explicitly
        (e.g. the policy used by :meth:`__setitem__` and :meth:`update`).
        Defaults to ``None``, which causes the *on_dup_kv* policy to match
        whatever *on_dup_val* policy is in effect.

    .. py:attribute:: fwdm

        Managed by bidict (you shouldn't need to touch this)
        but made public since we're consenting adults.

        The backing :class:`Mapping <collections.abc.Mapping>`
        storing the forward mapping data (*key* → *value*).

    .. py:attribute:: invm

        Managed by bidict (you shouldn't need to touch this)
        but made public since we're consenting adults.

        The backing :class:`Mapping <collections.abc.Mapping>`
        storing the inverse mapping data (*value* → *key*).

    .. py:attribute:: isinv

        Managed by bidict (you shouldn't need to touch this)
        but made public since we're consenting adults.

        :class:`bool` representing whether this bidict is the inverse of some
        other bidict which has already been created. If True, the meaning of
        :attr:`fwd_cls` and :attr:`inv_cls` is swapped. This enables
        the inverse of a bidict specifying a different :attr:`fwd_cls` and
        :attr:`inv_cls` to be passed back into its constructor such that
        the resulting copy has its :attr:`fwd_cls` and :attr:`inv_cls`
        set correctly.
    """

    on_dup_key = OVERWRITE
    on_dup_val = RAISE
    on_dup_kv = None
    fwd_cls = dict
    inv_cls = dict

    def __init__(self, *args, **kw):
        """Like dict's ``__init__``."""
        self.isinv = getattr(args[0], 'isinv', False) if args else False
        self.fwdm = self.inv_cls() if self.isinv else self.fwd_cls()
        self.invm = self.fwd_cls() if self.isinv else self.inv_cls()
        self._init_inv()  # lgtm [py/init-calls-subclass]
        self._hash = None
        if args or kw:
            self._update(True, self.on_dup_key, self.on_dup_val, self.on_dup_kv, *args, **kw)

    def _init_inv(self):
        inv = object.__new__(self.__class__)
        inv.isinv = not self.isinv
        inv.fwd_cls = self.inv_cls
        inv.inv_cls = self.fwd_cls
        inv.fwdm = self.invm
        inv.invm = self.fwdm
        inv.inv = self
        self.inv = inv

    def __repr__(self):
        tmpl = self.__class__.__name__ + '('
        if not self:
            return tmpl + ')'
        tmpl += '%r)'
        # If we have a truthy __reversed__ attribute, use an ordered repr.
        # (Python doesn't provide an Ordered or OrderedMapping ABC, else we'd
        # check that. Must use getattr rather than hasattr since __reversed__
        # may be set to None, which signifies non-ordered/-reversible.)
        ordered = bool(getattr(self, '__reversed__', False))
        delegate = list if ordered else dict
        return tmpl % delegate(iteritems(self))

    def __eq__(self, other):
        """Like :py:meth:`dict.__eq__`."""
        # This should be faster than using Mapping.__eq__'s implementation.
        return self.fwdm == other

    def __hash__(self):
        """
        Return the hash of this bidict from its contained items.

        Delegates to :meth:`compute_hash` on the first call,
        then caches the result to make future calls *O(1)*.
        """
        if self._hash is None:
            self._hash = self.compute_hash()
        return self._hash

    def compute_hash(self):
        """
        Use the pure Python implementation of Python's frozenset hashing
        algorithm from ``collections.Set._hash`` to compute the hash
        incrementally in constant space.
        """
        return ItemsView(self)._hash()  # pylint: disable=protected-access

    def _pop(self, key):
        val = self.fwdm.pop(key)
        del self.invm[val]
        return val

    def _clear(self):
        self.fwdm.clear()
        self.invm.clear()

    def _put(self, key, val, on_dup_key, on_dup_val, on_dup_kv):
        result = self._dedup_item(key, val, on_dup_key, on_dup_val, on_dup_kv)
        if result:
            self._write_item(key, val, *result)

    def _dedup_item(self, key, val, on_dup_key, on_dup_val, on_dup_kv):
        """
        Check *key* and *val* for any duplication in self.

        Handle any duplication as per the given duplication policies.

        (key, val) already present is construed as a no-op, not a duplication.

        If duplication is found and the corresponding duplication policy is
        *RAISE*, raise the appropriate error.

        If duplication is found and the corresponding duplication policy is
        *IGNORE*, return *None*.

        If duplication is found and the corresponding duplication policy is
        *OVERWRITE*, or if no duplication is found, return the dedup result
        *(isdupkey, isdupval, invbyval, fwdbykey)*.
        """
        if on_dup_kv is None:
            on_dup_kv = on_dup_val
        fwdm = self.fwdm
        invm = self.invm
        fwdbykey = fwdm.get(key, _MISS)
        invbyval = invm.get(val, _MISS)
        isdupkey = fwdbykey is not _MISS
        isdupval = invbyval is not _MISS
        if isdupkey and isdupval:
            if self._isdupitem(key, val, invbyval, fwdbykey):
                # (key, val) duplicates an existing item -> no-op.
                return
            # key and val each duplicate a different existing item.
            if on_dup_kv is RAISE:
                raise KeyAndValueDuplicationError(key, val)
            elif on_dup_kv is IGNORE:
                return
            # else on_dup_kv is OVERWRITE. Fall through to return on last line.
        elif isdupkey:
            if on_dup_key is RAISE:
                raise KeyDuplicationError(key)
            elif on_dup_key is IGNORE:
                return
            # else on_dup_key is OVERWRITE. Fall through to return on last line.
        elif isdupval:
            if on_dup_val is RAISE:
                raise ValueDuplicationError(val)
            elif on_dup_val is IGNORE:
                return
            # else on_dup_val is OVERWRITE. Fall through to return on last line.
        # else neither isdupkey nor isdupval.
        return isdupkey, isdupval, invbyval, fwdbykey

    @staticmethod
    def _isdupitem(key, val, oldkey, oldval):
        dup = oldkey == key
        assert dup == (oldval == val)
        return dup

    def _write_item(self, key, val, isdupkey, isdupval, oldkey, oldval):
        self.fwdm[key] = val
        self.invm[val] = key
        if isdupkey:
            del self.invm[oldval]
        if isdupval:
            del self.fwdm[oldkey]
        return key, val, isdupkey, isdupval, oldkey, oldval

    def _update(self, init, on_dup_key, on_dup_val, on_dup_kv, *args, **kw):
        if not args and not kw:
            return
        if on_dup_kv is None:
            on_dup_kv = on_dup_val
        rollbackonfail = not init or RAISE in (on_dup_key, on_dup_val, on_dup_kv)
        if rollbackonfail:
            return self._update_with_rollback(on_dup_key, on_dup_val, on_dup_kv, *args, **kw)
        _put = self._put
        for (key, val) in pairs(*args, **kw):
            _put(key, val, on_dup_key, on_dup_val, on_dup_kv)

    def _update_with_rollback(self, on_dup_key, on_dup_val, on_dup_kv, *args, **kw):
        """Update, rolling back on failure."""
        writes = []
        appendwrite = writes.append
        dedup_item = self._dedup_item
        write_item = self._write_item
        for (key, val) in pairs(*args, **kw):
            try:
                dedup_result = dedup_item(key, val, on_dup_key, on_dup_val, on_dup_kv)
            except DuplicationError:
                undo_write = self._undo_write
                for write in reversed(writes):
                    undo_write(*write)
                raise
            if dedup_result:
                write_result = write_item(key, val, *dedup_result)
                appendwrite(write_result)

    def _undo_write(self, key, val, isdupkey, isdupval, oldkey, oldval):
        fwdm = self.fwdm
        invm = self.invm
        if not isdupkey and not isdupval:
            self._pop(key)
            return
        if isdupkey:
            fwdm[key] = oldval
            invm[oldval] = key
            if not isdupval:
                del invm[val]
        if isdupval:
            invm[val] = oldkey
            fwdm[oldkey] = val
            if not isdupkey:
                del fwdm[key]

    def copy(self):
        """Like :py:meth:`dict.copy`."""
        # This should be faster than ``return self.__class__(self)`` because
        # it avoids unnecessary duplication checking.
        copy = object.__new__(self.__class__)
        copy.isinv = self.isinv
        copy.fwdm = self.fwdm.copy()
        copy.invm = self.invm.copy()
        copy._init_inv()  # pylint: disable=protected-access
        return copy

    __copy__ = copy
    __len__ = _proxied('__len__')
    __iter__ = _proxied('__iter__')
    __getitem__ = _proxied('__getitem__')
    values = _proxied('keys', attrname='inv')
    values.__doc__ = \
        "B.values() -> a set-like object providing a view on B's values.\n\n" \
        'Note that because values of a BidirectionalMapping are also keys\n' \
        'of its inverse, this returns a *KeysView* object rather than a\n' \
        '*ValuesView* object, conferring set-alike benefits.'
    if PY2:
        viewkeys = _proxied('viewkeys')

        viewvalues = _proxied('viewkeys', attrname='inv',
                              doc=values.__doc__.replace('values()', 'viewvalues()'))
        values.__doc__ = "Like dict's ``values``."

        # Use ItemsView here rather than proxying to fwdm.viewitems() so that
        # OrderedBidictBase (whose fwdm's values are nodes, not bare values)
        # can use it.
        viewitems = ItemsView



# -*- coding: utf-8 -*-
# Copyright 2017 Joshua Bronson. All Rights Reserved.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

u"""
Compatibility helpers.

    .. py:attribute:: PY2

        True iff running on Python 2.

    .. py:attribute:: PYPY

        True iff running on PyPy.

    .. py:attribute:: viewkeys

        viewkeys(D) → a set-like object providing a view on D's keys.

    .. py:attribute:: viewvalues

        viewvalues(D) → an object providing a view on D's values.

    .. py:attribute:: viewitems

        viewitems(D) → a set-like object providing a view on D's items.

    .. py:attribute:: iterkeys

        iterkeys(D) → an iterator over the keys of D.

    .. py:attribute:: itervalues

        itervalues(D) → an iterator over the values of D.

    .. py:attribute:: iteritems

        iteritems(D) → an iterator over the (key, value) items of D.

    .. py:attribute:: izip

        Alias for :func:`zip` on Python 3 / ``itertools.izip`` on Python 2.

"""

# pylint: disable-all

from operator import methodcaller
from platform import python_implementation
from sys import version_info

_compose = lambda f, g: lambda x: f(g(x))

PY2 = version_info[0] == 2
PYPY = python_implementation() == 'PyPy'

if PY2:
    assert version_info[1] > 6, 'Python >= 2.7 required'
    viewkeys = methodcaller('viewkeys')
    viewvalues = methodcaller('viewvalues')
    viewitems = methodcaller('viewitems')
    iterkeys = methodcaller('iterkeys')
    itervalues = methodcaller('itervalues')
    iteritems = methodcaller('iteritems')
    from itertools import izip
else:
    viewkeys = methodcaller('keys')
    viewvalues = methodcaller('values')
    viewitems = methodcaller('items')
    iterkeys = _compose(iter, viewkeys)
    itervalues = _compose(iter, viewvalues)
    iteritems = _compose(iter, viewitems)
    izip = zip



# -*- coding: utf-8 -*-
# Copyright 2017 Joshua Bronson. All Rights Reserved.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""Define bidict package metadata."""


def _get_version(fallback=u'0.0.0.version_not_found'):
    """Return bidict version via pkg_resources, or fallback if not found."""
    from pkg_resources import get_distribution, DistributionNotFound
    try:
        return get_distribution('bidict').version
    except DistributionNotFound:  # pragma: no cover
        return fallback


__author__ = u'Joshua Bronson'
__maintainer__ = u'Joshua Bronson'
__copyright__ = u'Copyright 2017 Joshua Bronson'
__email__ = u'jab@math.brown.edu'

# see ../docs/thanks.rst.inc
__credits__ = [i.strip() for i in u"""
Joshua Bronson, Michael Arntzenius, Francis Carr, Gregory Ewing, Raymond Hettinger, Jozef Knaperek,
Daniel Pope, Terry Reedy, David Turner, Tom Viner
""".split(u',')]

__description__ = u'Efficient, Pythonic bidirectional map implementation and related functionality'

__license__ = u'MPL 2.0'
__status__ = u'Beta'
__version__ = _get_version()



# -*- coding: utf-8 -*-
# Copyright 2017 Joshua Bronson. All Rights Reserved.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
Equality and hashing tests.
"""

from collections import Counter, Hashable, OrderedDict, Mapping, defaultdict
from itertools import product

import pytest
from bidict import FrozenOrderedBidict, OrderedBidict, bidict, namedbidict, frozenbidict


# pylint: disable=C0111
class DictSubcls(dict):
    pass


class OrderedBidictSubcls(OrderedBidict):
    pass


# pylint: disable=C0103
items = [('a', 1), ('b', 2)]  # use int values so makes sense with Counter
itemsreversed = list(reversed(items))

bidict_of_items = bidict(items)
frozenbidict_of_items = frozenbidict(items)
namedbidict_of_items = namedbidict('named', 'keys', 'vals')(items)
orderedbidict_of_items = OrderedBidict(items)
orderedbidict_of_itemsreversed = OrderedBidict(itemsreversed)
orderedbidictsubcls_of_items = OrderedBidictSubcls(items)
orderedbidictsubcls_of_itemsreversed = OrderedBidictSubcls(itemsreversed)
frozenorderedbidict_of_items = FrozenOrderedBidict(items)
frozenorderedbidict_of_itemsreversed = FrozenOrderedBidict(itemsreversed)
bidicts = (
    bidict_of_items,
    frozenbidict_of_items,
    namedbidict_of_items,
    orderedbidict_of_items,
    orderedbidict_of_itemsreversed,
    orderedbidictsubcls_of_items,
    orderedbidictsubcls_of_itemsreversed,
    frozenorderedbidict_of_items,
    frozenorderedbidict_of_itemsreversed,
)

dict_of_itemsreversed = dict(itemsreversed)
counter_of_itemsreversed = Counter(dict_of_itemsreversed)
defaultdict_of_itemsreversed = defaultdict(lambda x: None, itemsreversed)
dictsubcls_of_itemsreversed = DictSubcls(itemsreversed)
ordereddict_of_items = OrderedDict(items)
ordereddict_of_itemsreversed = OrderedDict(itemsreversed)
empty_dict = {}
not_a_mapping = 42
not_bidicts = (
    dict_of_itemsreversed,
    counter_of_itemsreversed,
    defaultdict_of_itemsreversed,
    dictsubcls_of_itemsreversed,
    ordereddict_of_items,
    ordereddict_of_itemsreversed,
    empty_dict,
    not_a_mapping,
)


@pytest.mark.parametrize('b, other', product(bidicts, bidicts + not_bidicts))
def test_eq_and_hash(b, other):
    if not isinstance(other, Mapping):
        assert b != other
    elif len(b) != len(other):
        assert b != other
    else:
        should_be_equal = dict(b) == dict(other)
        are_equal = b == other
        assert are_equal == should_be_equal
        both_hashable = isinstance(b, Hashable) and isinstance(other, Hashable)
        if are_equal and both_hashable:
            assert hash(b) == hash(other)

        if hasattr(b, 'equals_order_sensitive') and hasattr(other, '__reversed__'):
            are_equal_ordered = b.equals_order_sensitive(other)
            should_be_equal_ordered = OrderedDict(b) == OrderedDict(other)
            assert are_equal_ordered == should_be_equal_ordered



# -*- coding: utf-8 -*-
# Copyright 2017 Joshua Bronson. All Rights Reserved.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""Provides :class:`_Marker` for representing singletons."""


class _Marker(object):
    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return '<%s>' % self.name  # pragma: no cover



# -*- coding: utf-8 -*-
# Copyright 2017 Joshua Bronson. All Rights Reserved.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""Implements :class:`bidict.OrderedBidict` and :class:`bidict.FrozenOrderedBidict`."""

from collections import Mapping

from ._bidict import bidict
from ._frozen import frozenbidict
from ._marker import _Marker
from ._miss import _MISS
from .compat import iteritems, izip


_PRV = 1
_NXT = 2
_END = _Marker('END')


class FrozenOrderedBidict(frozenbidict):
    u"""
    Frozen ordered bidict. Base class for :class:`OrderedBidict`.

    .. py:attribute:: sntl

        Managed by bidict (you shouldn't need to touch this)
        but made public since we're consenting adults.

        Sentinel node for the backing circular doubly-linked list of
        [*data*, *previous*, *next*] nodes storing the ordering.
        The *data* contained in each node is (*key_or_val*, *val_or_key*).

    .. py:attribute:: fwdm

        Managed by bidict (you shouldn't need to touch this)
        but made public since we're consenting adults.

        Backing dict storing the forward mapping data (*key* → *node*).

    .. py:attribute:: invm

        Managed by bidict (you shouldn't need to touch this)
        but made public since we're consenting adults.

        Backing dict storing the inverse mapping data (*value* → *node*).
    """

    # Make explicit that FrozenOrderedBidict does not support overriding these:
    fwd_cls = None
    inv_cls = None

    def __init__(self, *args, **kw):
        """Like :meth:`collections.OrderedDict.__init__`."""
        # pylint: disable=super-init-not-called
        self.isinv = getattr(args[0], 'isinv', False) if args else False
        self.fwdm = {}
        self.invm = {}
        self.sntl = self._make_sentinel()
        self._hash = None
        self._init_inv()
        if args or kw:
            self._update(True, self.on_dup_key, self.on_dup_val, self.on_dup_kv, *args, **kw)

    @staticmethod
    def _make_sentinel():
        sntl = []
        sntl[:] = [_END, sntl, sntl]
        return sntl

    def _init_inv(self):
        super(FrozenOrderedBidict, self)._init_inv()
        self.inv.sntl = self.sntl

    # Can't reuse frozenbidict.copy since we have different internal structure.
    def copy(self):
        """Like :meth:`collections.OrderedDict.copy`."""
        # This should be faster than ``return self.__class__(self)`` because
        # it avoids unnecessary duplication checking.
        copy = object.__new__(self.__class__)
        copy.isinv = self.isinv
        sntl = self._make_sentinel()
        fwdm = {}
        invm = {}
        cur = sntl
        nxt = sntl[_NXT]
        for (key, val) in iteritems(self):
            nxt = [(key, val), cur, sntl]
            cur[_NXT] = fwdm[key] = invm[val] = nxt
            cur = nxt
        sntl[_PRV] = nxt
        copy.sntl = sntl
        copy.fwdm = fwdm
        copy.invm = invm
        copy._init_inv()  # pylint: disable=protected-access
        return copy

    __copy__ = copy

    def _clear(self):
        self.fwdm.clear()
        self.invm.clear()
        self.sntl[:] = [_END, self.sntl, self.sntl]

    @staticmethod
    def _get_other(nodedata, key_or_val):
        first, second = nodedata
        if key_or_val == first:
            return second
        assert key_or_val == second
        return first

    def __getitem__(self, key):
        nodefwd = self.fwdm[key]
        datafwd = nodefwd[0]
        val = self._get_other(datafwd, key)
        nodeinv = self.invm[val]
        assert nodeinv is nodefwd
        return val

    def _pop(self, key):
        nodefwd = self.fwdm.pop(key)
        datafwd, prv, nxt = nodefwd
        val = self._get_other(datafwd, key)
        nodeinv = self.invm.pop(val)
        assert nodeinv is nodefwd
        prv[_NXT] = nxt
        nxt[_PRV] = prv
        del nodefwd[:]
        return val

    @staticmethod
    def _isdupitem(key, val, nodeinv, nodefwd):  # pylint: disable=arguments-differ
        """Return whether (key, val) duplicates an existing item."""
        return nodeinv is nodefwd

    # pylint: disable=arguments-differ
    def _write_item(self, key, val, isdupkey, isdupval, nodeinv, nodefwd):
        fwdm = self.fwdm
        invm = self.invm
        if not isdupkey and not isdupval:
            sntl = self.sntl
            last = sntl[_PRV]
            node = [(key, val), last, sntl]
            sntl[_PRV] = last[_NXT] = fwdm[key] = invm[val] = node
            oldkey = oldval = _MISS
        elif isdupkey and isdupval:
            datafwd = nodefwd[0]
            oldval = self._get_other(datafwd, key)
            datainv, invprv, invnxt = nodeinv
            oldkey = self._get_other(datainv, val)
            assert oldkey != key
            assert oldval != val
            # We have to collapse nodefwd and nodeinv into a single node, i.e. drop one of them.
            # Drop nodeinv, so that the item with the same key is the one overwritten in place.
            invprv[_NXT] = invnxt
            invnxt[_PRV] = invprv
            # Don't remove nodeinv's references to its neighbors since
            # if the update fails, we'll need them to undo this write.
            # Python's garbage collector should still be able to detect when
            # nodeinv is garbage and reclaim the memory.
            # Update fwdm and invm.
            tmp = fwdm.pop(oldkey)
            assert tmp is nodeinv
            tmp = invm.pop(oldval)
            assert tmp is nodefwd
            fwdm[key] = invm[val] = nodefwd
            # Update nodefwd with new item.
            nodefwd[0] = (key, val)
        elif isdupkey:
            datafwd = nodefwd[0]
            oldval = self._get_other(datafwd, key)
            oldkey = _MISS
            oldnodeinv = invm.pop(oldval)
            assert oldnodeinv is nodefwd
            invm[val] = nodefwd
            node = nodefwd
        elif isdupval:
            datainv = nodeinv[0]
            oldkey = self._get_other(datainv, val)
            oldval = _MISS
            oldnodefwd = fwdm.pop(oldkey)
            assert oldnodefwd is nodeinv
            fwdm[key] = nodeinv
            node = nodeinv
        if isdupkey ^ isdupval:
            node[0] = (key, val)
        return key, val, isdupkey, isdupval, nodeinv, nodefwd, oldkey, oldval

    # pylint: disable=arguments-differ
    def _undo_write(self, key, val, isdupkey, isdupval, nodeinv, nodefwd, oldkey, oldval):  # lgtm
        fwdm = self.fwdm
        invm = self.invm
        if not isdupkey and not isdupval:
            self._pop(key)
        elif isdupkey and isdupval:
            # datainv was never changed so should still have the original item.
            datainv, invprv, invnxt = nodeinv
            assert datainv == (val, oldkey) or datainv == (oldkey, val)
            # Restore original items.
            nodefwd[0] = (key, oldval)
            invprv[_NXT] = invnxt[_PRV] = nodeinv
            fwdm[oldkey] = invm[val] = nodeinv
            invm[oldval] = fwdm[key] = nodefwd
        elif isdupkey:
            nodefwd[0] = (key, oldval)
            tmp = invm.pop(val)
            assert tmp is nodefwd
            invm[oldval] = nodefwd
            assert fwdm[key] is nodefwd
        elif isdupval:
            nodeinv[0] = (oldkey, val)
            tmp = fwdm.pop(key)
            assert tmp is nodeinv
            fwdm[oldkey] = nodeinv
            assert invm[val] is nodeinv

    def __iter__(self, reverse=False):
        """Like :meth:`collections.OrderedDict.__iter__`."""
        fwdm = self.fwdm
        sntl = self.sntl
        cur = sntl[_PRV if reverse else _NXT]
        while cur is not sntl:
            data, prv, nxt = cur
            korv = data[0]
            node = fwdm.get(korv)
            key = korv if node is cur else data[1]
            yield key
            cur = prv if reverse else nxt

    def __reversed__(self):
        """Like :meth:`collections.OrderedDict.__reversed__`."""
        # python2 lacks `yield from` syntax
        for key in self.__iter__(reverse=True):
            yield key

    def __eq__(self, other, order_sensitive=False):
        """Check for equality with ``other`` as per ``order_sensitive``."""
        if not isinstance(other, Mapping):
            return NotImplemented
        if len(self) != len(other):
            return False
        if order_sensitive:
            return all(i == j for (i, j) in izip(iteritems(self), iteritems(other)))
        return all(self.get(k, _MISS) == v for (k, v) in iteritems(other))

    def equals_order_sensitive(self, other):
        """Check equality with other with order sensitivity."""
        return self.__eq__(other, order_sensitive=True)

    # frozenbidict.__hash__ is also correct for ordered bidicts:
    # The value is derived from all contained items and insensitive to their order.
    # If an ordered bidict "O" is equal to a mapping, its unordered counterpart "U" is too.
    # Since U1 == U2 => hash(U1) == hash(U2), then if O == U1, hash(O) must equal hash(U1).
    __hash__ = frozenbidict.__hash__  # Must set explicitly, __hash__ is never inherited.


class OrderedBidict(FrozenOrderedBidict, bidict):
    """Mutable bidict type that maintains items in insertion order."""

    __hash__ = None  # since this class is mutable. explicit > implicit.

    def popitem(self, last=True):  # pylint: disable=arguments-differ
        """Like :meth:`collections.OrderedDict.popitem`."""
        if not self:
            raise KeyError(self.__class__.__name__ + ' is empty')
        key = next((reversed if last else iter)(self))
        val = self._pop(key)
        return key, val

    def move_to_end(self, key, last=True):
        """Like :meth:`collections.OrderedDict.move_to_end`."""
        node = self.fwdm[key]
        _, prv, nxt = node
        prv[_NXT] = nxt
        nxt[_PRV] = prv
        sntl = self.sntl
        if last:
            last = sntl[_PRV]
            node[_PRV] = last
            node[_NXT] = sntl
            sntl[_PRV] = last[_NXT] = node
        else:
            frst = sntl[_NXT]
            node[_PRV] = sntl
            node[_NXT] = frst
            sntl[_NXT] = frst[_PRV] = node



# -*- coding: utf-8 -*-
# Copyright 2017 Joshua Bronson. All Rights Reserved.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
Test bidict metadata.
"""

import bidict


# pylint: disable=all
# (otherwise pylint flags duplicate code in bidict/metadata.py)
META_ATTRS = [
    '__author__',
    '__maintainer__',
    '__copyright__',
    '__email__',
    '__credits__',
    '__description__',
    '__license__',
    '__status__',
    '__version__',
]


def test_metadata():
    """Ensure bidict has expected metadata attributes."""
    for i in META_ATTRS:
        assert getattr(bidict, i)



# -*- coding: utf-8 -*-
# Copyright 2017 Joshua Bronson. All Rights Reserved.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
Efficient, Pythonic bidirectional map implementation and related functionality.

See https://bidict.readthedocs.io for comprehensive documentation.

.. :copyright: (c) 2017 Joshua Bronson.
.. :license: MPLv2. See LICENSE for details.

"""

from ._abc import BidirectionalMapping
from ._bidict import bidict
from ._dup import DuplicationPolicy, IGNORE, OVERWRITE, RAISE
from ._exc import (
    BidictException, DuplicationError,
    KeyDuplicationError, ValueDuplicationError, KeyAndValueDuplicationError)
from ._frozen import frozenbidict
from ._named import namedbidict
from ._ordered import FrozenOrderedBidict, OrderedBidict
from .metadata import (
    __author__, __maintainer__, __copyright__, __email__, __credits__,
    __license__, __status__, __description__, __version__)
from .util import pairs, inverted


__all__ = (
    '__author__',
    '__maintainer__',
    '__copyright__',
    '__email__',
    '__credits__',
    '__license__',
    '__status__',
    '__description__',
    '__version__',
    'BidirectionalMapping',
    'BidictException',
    'DuplicationPolicy',
    'IGNORE',
    'OVERWRITE',
    'RAISE',
    'DuplicationError',
    'KeyDuplicationError',
    'ValueDuplicationError',
    'KeyAndValueDuplicationError',
    'frozenbidict',
    'bidict',
    'namedbidict',
    'FrozenOrderedBidict',
    'OrderedBidict',
    'pairs',
    'inverted',
)



# -*- coding: utf-8 -*-
# Copyright 2017 Joshua Bronson. All Rights Reserved.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""Provides bidict ABCs."""

from collections import Mapping

from .compat import iteritems


class BidirectionalMapping(Mapping):  # pylint: disable=abstract-method
    """Abstract base class for bidirectional mappings.
    Extends :class:`collections.abc.Mapping`.

    .. py:attribute:: inv

        The inverse mapping.

    .. py:attribute:: _subclsattrs

        The attributes that :attr:`__subclasshook__` checks for to determine
        whether a class is a subclass of :class:`BidirectionalMapping`.
    """

    __slots__ = ()

    inv = NotImplemented

    def __inverted__(self):
        """Get an iterator over the items in :attr:`inv`."""
        return iteritems(self.inv)

    _subclsattrs = frozenset({
        'inv', '__inverted__',
        # see "Mapping" in the table at
        # https://docs.python.org/3/library/collections.abc.html#collections-abstract-base-classes
        '__getitem__', '__iter__', '__len__',  # abstract methods
        '__contains__', 'keys', 'items', 'values', 'get', '__eq__', '__ne__',  # mixin methods
    })

    @classmethod
    def __subclasshook__(cls, C):  # noqa: N803 ("argument name should be lowercase")
        # Standard to use "C" for this arg in __subclasshook__, e.g.:
        # https://github.com/python/cpython/blob/d505a2/Lib/_collections_abc.py#L93
        """Check if C provides all the attributes in :attr:`_subclsattrs`.

        Causes conforming classes to be virtual subclasses automatically.
        """
        if cls is not BidirectionalMapping:  # lgtm [py/comparison-using-is]
            return NotImplemented
        mro = getattr(C, '__mro__', None)
        if mro is None:
            return NotImplemented
        return all(any(B.__dict__.get(i) for B in mro) for i in cls._subclsattrs)



# -*- coding: utf-8 -*-
# Copyright 2017 Joshua Bronson. All Rights Reserved.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
setup.py
"""

from io import open  # pylint: disable=redefined-builtin

from setuptools import setup

# Fetch bidict's package metadata from bidict/metadata.py.
# Must use exec(open(...)) because we haven't been installed yet.
METADATA = {}
exec(open('bidict/metadata.py').read().encode('utf8'), METADATA)  # pylint: disable=exec-used

try:
    LONG_DESCRIPTION = open('README.rst').read().replace(
        ':doc:', '(doc:)'  # :doc: breaks rendering on PyPI
    )
except:  # noqa; pylint: disable=bare-except
    LONG_DESCRIPTION = 'See https://bidict.readthedocs.io'

SETUP_REQ = [
    'pytest-runner',
    'setuptools_scm',
]

TESTS_REQ = [
    'hypothesis==3.38.0',
    'hypothesis-pytest==0.19.0',
    'py==1.5.2',
    'pytest==3.2.5',
    'pytest-benchmark==3.1.1',
    'sortedcollections==0.5.3',
    'sortedcontainers==1.5.7',
]

COVERAGE_REQ = [
    'coverage==4.4.2',
    'pytest-cov==2.5.1',
]

DEV_REQ = SETUP_REQ + TESTS_REQ + COVERAGE_REQ + [
    'Sphinx==1.6.5',
    'flake8==3.5.0',
    'pre-commit==1.4.1',
    'pydocstyle==2.1.1',
    'pylint==1.7.4',
    'tox==2.9.1',
]

setup(
    name='bidict',
    use_scm_version=True,
    author=METADATA['__author__'],
    author_email=METADATA['__email__'],
    description=METADATA['__description__'],
    long_description=LONG_DESCRIPTION,
    keywords='dict, dictionary, mapping, datastructure, '
             'bimap, bijection, bijective, injective, inverse, reverse, '
             'bidirectional, two-way, 2-way',
    url='https://github.com/jab/bidict',
    license=METADATA['__license__'],
    packages=['bidict'],
    zip_safe=False,  # Don't zip. (We're zip-safe but prefer not to.)
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    setup_requires=SETUP_REQ,
    tests_require=TESTS_REQ,
    extras_require=dict(
        test=TESTS_REQ,
        coverage=COVERAGE_REQ,
        dev=DEV_REQ,
    ),
)



# -*- coding: utf-8 -*-
# Copyright 2017 Joshua Bronson. All Rights Reserved.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""Provides bidict duplication behaviors."""

from ._marker import _Marker


class DuplicationPolicy(_Marker):
    """
    Provide RAISE, OVERWRITE, and IGNORE duplication policies.

    .. py:attribute:: RAISE

        Raise an exception when a duplication is encountered.

    .. py:attribute:: OVERWRITE

        Overwrite an existing item when a duplication is encountered.

    .. py:attribute:: IGNORE

        Keep the existing item and ignore the new item when a duplication is
        encountered.
    """


DuplicationPolicy.RAISE = RAISE = DuplicationPolicy('RAISE')
DuplicationPolicy.OVERWRITE = OVERWRITE = DuplicationPolicy('OVERWRITE')
DuplicationPolicy.IGNORE = IGNORE = DuplicationPolicy('IGNORE')



# -*- coding: utf-8 -*-
# Copyright 2017 Joshua Bronson. All Rights Reserved.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""Provides all bidict exceptions."""


class BidictException(Exception):
    """Base class for bidict exceptions."""


class DuplicationError(BidictException):
    """Base class for exceptions raised when uniqueness is violated."""


class KeyDuplicationError(DuplicationError):
    """Raised when a given key is not unique."""


class ValueDuplicationError(DuplicationError):
    """Raised when a given value is not unique."""


class KeyAndValueDuplicationError(KeyDuplicationError, ValueDuplicationError):
    """
    Raised when a given item's key and value are not unique.

    That is, its key duplicates that of another item,
    and its value duplicates that of a different other item.
    """



# -*- coding: utf-8 -*-
# Copyright 2017 Joshua Bronson. All Rights Reserved.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
Test that if foreign code provides a class that conforms to
BidirectionalMapping's interface, it is automatically a subclass.
"""

from bidict import BidirectionalMapping


class MyBidirectionalMapping(dict):
    """Dummy type implementing the BidirectionalMapping interface."""
    def __inverted__(self):
        for (key, val) in self.items():
            yield (val, key)

    @property
    def inv(self):
        """Like :attr:`bidict.bidict.inv`."""
        return MyBidirectionalMapping(self.__inverted__())


class OldStyleClass:
    """In Python 2 this is an old-style class (not derived from object)."""


def test_bidi_mapping_subclasshook():
    """Ensure issubclass(foo, BidirectionalMapping) works as expected."""
    assert issubclass(MyBidirectionalMapping, BidirectionalMapping)
    assert not issubclass(dict, BidirectionalMapping)
    # Make sure this works with old-style classes as expected.
    assert not issubclass(OldStyleClass, BidirectionalMapping)
